import sqlite3

from getpass import getuser
from importlib import import_module
from izzylib import BaseConfig, Path

from .result import Result
from .row import Row
from .session import Session


class Config(BaseConfig):
	def __init__(self, **kwargs):
		super().__init__(
			appname = 'IzzyLib SQL Client',
			type = 'sqlite',
			module = None,
			module_name = None,
			tables = {},
			row_classes = {},
			session_class = Session,
			result_class = Result,
			host = 'localhost',
			port = 0,
			database = None,
			username = getuser(),
			password = None,
			minconnections = 4,
			maxconnections = 25,
			engine_args = {},
			auto_trans = True,
			connect_function = None,
			autocommit = False
		)

		for k, v in kwargs.items():
			self[k] = v

		if not self.database:
			if self.type == 'sqlite':
				self.database = ':memory:'

			else:
				raise ValueError('Missing database name')

		if not self.port:
			if self.type == 'postgresql':
				self.port = 5432

			elif self.type == 'mysql':
				self.port = 3306

		if not self.module and not self.connect_function:
			if self.type == 'sqlite':
				self.module = sqlite3
				self.module_name = 'sqlite3'

			elif self.type == 'postgresql':
				for mod in ['pg8000.dbapi', 'pgdb', 'psycopg2']:
					try:
						self.module = import_module(mod)
						self.module_name = mod
						break

					except ImportError:
						pass

			elif self.type == 'mysql':
				try:
					self.module = import_module('mysql.connector')
					self.module_name = 'mysql.connector'
				except ImportError:
					pass

			if not self.module:
				raise ImportError(f'Cannot find module for "{self.type}"')

			self.module.paramstyle = 'qmark'

		if self.type == 'sqlite':
			self.engine_args['check_same_thread'] = False


	@property
	def dbargs(self):
		return {key: self[key] for key in ['host', 'port', 'database', 'username', 'password']}


	def parse_value(self, key, value):
		if key == 'type':
			if value not in ['sqlite', 'postgresql', 'mysql', 'mssql']:
				raise ValueError(f'Invalid database type: {value}')

		if key == 'port':
			if not isinstance(value, int):
				raise TypeError('Port is not an integer')

		if key == 'row_classes':
			for row_class in value.values():
				if not issubclass(row_class, Row):
					raise TypeError(f'Row classes must be izzylib.sql2.row.Row, not {row_class.__name__}')

		return value
