import itertools

from izzylib import DotDict, ObjectBase, Path, class_name, izzylog
from threading import Thread

from .config import Config
from .connection import Connection
from .row import Row
from .table import DbTables
from .types import Types

from .exceptions import (
	MaxConnectionsError,
	NoTableLayoutError,
	NoConnectionError
)


class Database(ObjectBase):
	def __init__(self, autoconnect=True, app=None, **kwargs):
		super().__init__(
			cfg = Config(**kwargs),
			tables = DbTables(self),
			cache = DotDict(),
			types = Types(self),
			connections = [],
			app = app,
			callbacks = ObjectBase(
				insert = DotDict(),
				update = DotDict(),
				remove = DotDict(),
				readonly_props = True
			),
			readonly_props = True
		)

		self.types._setup_types()

		if self.cfg.tables:
			self.load_tables(self.cfg.tables)

		if autoconnect:
			self.connect()


	@property
	def connection(self):
		if not len(self.connections):
			return self.new_connection()

		if len(self.connections) < self.cfg.minconnections:
			return self.new_connection()

		for conn in self.connections:
			if not len(conn.sessions):
				return conn

		if len(self.connections) < self.cfg.maxconnections:
			return self.new_connection()

		conns = {(conn, len(conn.sessions)) for conn in self.connections}
		return min(conns, key=lambda x: x[1])[0]


	@property
	def session(self):
		return self.connection.session


	## decorator to register a function as a callback
	def callback(self, action, table, *args, thread=True, **kwargs):
		assert action in ['insert', 'update', 'remove']

		def wrapper(func):
			self.register_callback(action, table, func, *args, thread=thread, **kwargs)
			return func

		return wrapper


	def close_connection(self, conn):
		conn.close_sessions()
		conn.disconnect()

		if not conn.conn:
			try: self.connections.remove(conn)
			except: pass


	def connect(self):
		for _ in itertools.repeat(None, self.cfg.minconnections):
			self.connection

		with self.session as s:
			self.load_tables(s.table_layout())


	def create_database(self, tables=None):
		if self.cfg.type == 'postgresql':
			with self.new_predb().session as s:
				if not s.raw_execute('SELECT datname FROM pg_database WHERE datname = ?', [self.cfg.database]).fetchone():
					s.raw_execute(f'CREATE DATABASE {self.cfg.database}')

		elif self.cfg.type != 'sqlite':
			raise NotImplementedError(f'Database type not supported yet: {self.cfg.type}')

		self.create_tables(tables or self.cfg.tables)


	def create_tables(self, tables=None):
		if self.tables.empty:
			if not tables:
				raise ValueError('No table layout provided')

			self.load_tables(tables)

		with self.session as s:
			for table in self.tables.names:
				s.execute(self.tables.compile_table(table, self.cfg.type))

			s.commit()


	def disconnect(self):
		for conn in self.connections:
			conn.disconnect()

		self.connections = []
		self.cache.clear()


	def drop_database(self, database):
		if self.cfg.type == 'sqlite':
			izzylog.verbose('drop_database not needed for SQLite')
			return

		with self.session as s:
			if self.cfg.type == 'postgresql':
				s.raw_execute(f'DROP DATABASE {database}')

			else:
				raise NotImplementedError(f'Database type not supported yet: {self.cfg.type}')


	def get_row_class(self, name):
		return self.cfg.row_classes.get(name, Row)


	def load_tables(self, tables=None):
		if tables != None:
			self.tables.load_tables(tables)

		else:
			with self.session as s:
				self.tables.load_tables(s.table_layout())

		return self.tables


	def new_connection(self):
		if len(self.connections) >= self.cfg.maxconnections:
			raise MaxConnectionsError('Too many connections')

		conn = Connection(self)
		conn.connect()
		self.connections.append(conn)

		return conn


	def new_predb(self, database='postgres'):
		dbconfig = Config(**self.cfg)
		dbconfig['database'] = database
		dbconfig['autocommit'] = True

		return Database(**dbconfig)


	def register_callback(self, action, table, callback, *args, thread=False, **kwargs):
		assert action in ['insert', 'update', 'remove']

		self.callbacks[action][callback] = DotDict(
			table = table,
			args = args,
			kwargs = kwargs,
			thread = thread
		)


	def run_callback(self, action, table, rows):
		assert action in ['insert', 'update', 'remove']

		for func, data in self.callbacks[action].items():
			if table != data.table:
				continue

			if data.thread:
				Thread(target=func, args=(rows, *data.args), kwargs=data.kwargs).start()
				return

			func(rows, *data.args, **data.kwargs)


	def set_row_class(self, name, row_class):
		if not issubclass(row_class, Row):
			raise TypeError(f'Row classes must be izzylib_sql.row.Row, not {row_class.__name__}')

		self.cfg.row_classes[name] = row_class


	def unregister_all_callbacks(self):
		for action in ['insert', 'update', 'remove']:
			self.callbacks[action].clear()


	def unregister_callback(self, action, callback):
		assert action in ['insert', 'update', 'remove']

		self.callbacks[action].pop(callback, None)


	def vacuum(self):
		self.connection.execute('VACUUM')
