import json, traceback

from pathlib import Path as PyPath
from izzylib import DotDict, Path, convert_to_boolean, izzylog, random_str

from .exceptions import NoTransactionError, UpdateAllRowsError
from .result import Result
from .row import Row
from .statements import Count, Delete, Insert, Select, Statement, Update
from .table import SessionTables, DbColumn


class Session:
	def __init__(self, conn):
		self.db = conn.db
		self.cfg = conn.db.cfg
		self.conn = conn
		self.sid = random_str()
		self.tables = SessionTables(self)

		self.cursor = conn.cursor()
		self.trans = False

		if self.db.tables.empty:
			self.db.load_tables(self.table_layout())

		self.__setup__()


	def __enter__(self):
		return self


	def __exit__(self, exctype, excvalue, traceback):
		if traceback:
			self.rollback()

		else:
			self.commit()


	def __setup__(self):
		pass


	@property
	def cache(self):
		return self.db.cache


	def close():
		if not self.cursor:
			return

		self.conn.close_session(self)


	def _parse_data(self, action, table, kwargs):
		data = {}

		if self.db.tables:
			for key, value in kwargs.items():
				try:
					coltype = self.db.tables[table][key].type

				except KeyError:
					data[key] = value
					continue

				parser = self.db.types.get_type(coltype)

				try:
					data[key] = parser(action, self.cfg.type, value)
				except Exception as e:
					izzylog.error(f'Failed to parse data from the table "{table}": {key} = {value}')
					izzylog.debug(f'Parser: {parser}, Type: {coltype}')
					raise e from None

		else:
			data = kwargs

		return data


	def dump_database(self, path):
		import sqlparse
		path = Path(path)

		with path.open('w') as fd:
			line = '\n\n'.join(list(self.conn.iterdump())[1:-1])
			fd.write(sqlparse.format(line,
				reindent = False,
				keyword_case = 'upper',
			))

	def dump_database2(self, path):
		path = Path(path)

		with path.open('w') as fd:
			fd.write('\n\n'.join(list(self.conn.iterdump())[1:-1]))


	def begin(self):
		if self.trans or self.cfg.autocommit:
			return

		self.execute('BEGIN')
		self.trans = True


	def commit(self):
		if not self.trans:
			return

		self.execute('COMMIT')
		self.trans = False


	def rollback(self):
		if not self.trans:
			return

		self.execute('ROLLBACK')
		self.trans = False


	def raw_execute(self, string, values=None):
		if type(string) == Path:
			string = string.read()

		elif type(string) == PyPath:
			with string.open() as fd:
				string = fd.read()

		if values:
			self.cursor.execute(string, values)

		else:
			self.cursor.execute(string)

		return self.cursor


	def execute(self, string, *values):
		if isinstance(string, Statement):
			raise TypeError('String must be a str not a Statement')

		action = string.split()[0].upper()

		if not self.trans and action in ['CREATE', 'INSERT', 'UPDATE', 'UPSERT', 'DROP', 'DELETE', 'ALTER']:
			if self.cfg.auto_trans:
				self.begin()

			else:
				raise NoTransactionError(f'Command not supported outside a transaction: {action}')

		try:
			self.raw_execute(string, values)

		except Exception as e:
			if type(e).__name__ in ['DatabaseError', 'OperationalError']:
				print(string, values)

			raise e from None

		return Result(self)


	def run(self, query):
		result = self.execute(query.compile(self.cfg.type), *query.values)

		if type(query) == Count:
			return list(result.one().values())[0]

		result.set_table(query.table)
		return result


	def count(self, table, **kwargs):
		if self.db.tables and table not in self.db.tables:
			raise KeyError(f'Table does not exist: {table}')

		return self.run(Count(table, **kwargs))


	def fetch(self, table, orderby=None, orderdir='ASC', limit=None, offset=None, **kwargs):
		if self.db.tables and table not in self.db.tables:
			raise KeyError(f'Table does not exist: {table}')

		query = Select(table, **kwargs)

		if orderby:
			query.order(orderby, orderdir)

		if limit:
			query.limit(limist)

		if offset:
			query.offset(offset)

		return self.run(query)


	def insert(self, table, **kwargs):
		if self.db.tables and table not in self.db.tables:
			raise KeyError(f'Table does not exist: {table}')

		rows = self.run(Insert(table, **self._parse_data('serialize', table, kwargs))).all()

		if not rows:
			rows = self.fetch(table, **kwargs).all()

		self.db.run_callback('insert', table, rows)
		return rows


	def update(self, table, data, **kwargs):
		query = Update(table, **data)

		for pair in kwargs.items():
			query.where(*pair)

		# not sure if this should be kept
		if not query._where:
			raise UpdateAllRowsError(f'Refusing to update all rows in table: {table}')

		rows = self.run(query).all()

		if not rows:
			rows = self.fetch(table, **data).all()

		self.db.run_callback('update', table, rows)
		return rows


	def remove(self, table, **kwargs):
		if self.db.tables and table not in self.db.tables:
			raise KeyError(f'Table does not exist: {table}')

		rows = self.fetch(table, **kwargs).all()

		self.run(Delete(table, **self._parse_data('serialize', table, kwargs))).all()
		self.db.run_callback('remove', table, rows)

		return rows


	def update_row(self, row, **kwargs):
		if not row.table:
			raise ValueError('Row not associated with a table')

		rows = self.update(row.table, kwargs, id=row.id)

		try:
			return rows[0]

		except:
			return


	def remove_row(self, row):
		if not row.table:
			raise ValueError('Row not associated with a table')

		rows = self.remove(row.table, id=row.id)

		try:
			return rows[0]

		except:
			return


	def create_table(self, table):
		self.execute(self.tables[table].compile(self.cfg.type))


	def drop_table(self, table):
		self.execute(f'DROP TABLE {table}')


	def alter_table(self, table):
		return AlterTable(self, table)


	def create_tables(self, tables=None):
		if tables:
			self.db.load_tables(tables)

		if not self.tables:
			raise NoTableLayoutError('No table layout available')

		for table in self.db.tables.keys():
			self.execute(self.db.tables.compile_table(table, self.db.cfg.type))


	def truncate(self, table):
		if self.cfg.type == 'sqlite':
			self.execute(f"DELETE FROM '{table}'")

		elif self.cfg.type in ['mysql', 'postgresql']:
			self.execute(f"TRUNCATE TABLE '{table}")

		else:
			raise NotImplementedError(f'Unsupported database type: {self.cfg.type}')


	def table_layout(self):
		tables = {}

		if self.cfg.type == 'sqlite':
			rows = self.execute("SELECT name, sql FROM sqlite_master WHERE type = 'table' AND name NOT LIKE 'sqlite_%'")

			for row in rows:
				name = row.name
				tables[name] = {}
				fkeys = {fkey['from']: f'{fkey.table}.{fkey["to"]}' for fkey in self.execute(f'PRAGMA foreign_key_list({name})')}
				columns = [col for col in self.execute(f'PRAGMA table_info({name})')]

				unique_list = parse_unique(row.sql)

				for column in columns:
					tables[name][column.name] = dict(
						type = column.type.upper(),
						nullable = not column.notnull,
						default = parse_default(column.dflt_value),
						primary_key = bool(column.pk),
						foreign_key = fkeys.get(column.name),
						unique = column.name in unique_list
					)

		elif self.cfg.type == 'postgresql':
			for row in self.execute("SELECT * FROM information_schema.columns WHERE table_schema not in ('information_schema', 'pg_catalog') ORDER BY table_schema, table_name, ordinal_position"):
				table = row.table_name
				column = row.column_name

				if not tables.get(table):
					tables[table] = {}

				if not tables[table].get(column):
					tables[table][column] = {}

				tables[table][column] = dict(
					type = row.data_type.upper(),
					nullable = convert_to_boolean(row.is_nullable),
					default = row.column_default if row.column_default and not row.column_default.startswith('nextval') else None,
					primary_key = None,
					foreign_key = None,
					unique = None
				)

		else:
			raise NotImplementedError(f'Unsupported database type: {self.cfg.type}')

		return tables


class AlterTable:
	def __init__(self, session, table):
		self.session = session
		self.table = table


	def _execute(self, sql):
		return self.session.execute(f"ALTER TABLE '{self.table}' {sql}")


	def rename_table(self, new_name):
		return self._execute(f"RENAME TO '{new_name}'")


	def rename_column(self, column, new_name):
		return self._execute(f"RENAME COLUMN '{column}' TO '{new_name}'")


	def add_column(self, *args, dbtype='sqlite', **kwargs):
		col = DbColumn(dbtype, *args, **kwargs)

		if col.nullable and not col.default:
			raise ValueError('A default value must be provided for nullable columns')

		if col.foreign_key and col.default != None:
			raise ValueError('A default value of "None" must be used if a foreign key is set')

		return self._execute(f"ADD COLUMN {col.compile()}")


	def drop_column(self, column):
		return self._execute(f"DROP TABLE '{column}'")


def parse_unique(sql):
	unique_list = []

	try:
		for raw_line in sql.splitlines():
			if 'UNIQUE' not in raw_line:
				continue

			for line in raw_line.replace('UNIQUE', '').replace('(', '').replace(')', '').split(','):
				line = line.strip()

				if line:
					unique_list.append(line)

	except IndexError:
		pass

	return unique_list


def parse_default(value):
	if value == None:
		return

	if value.startswith("'") and value.endswith("'"):
		value = value[1:-1]

	else:
		try:
			value = int(value)

		except ValueError:
			pass

	return value
