from izzylib import DotDict

from . import SQLITE_VERSION
from .table import DbColumn


Comparison = DotDict(
	LESS = lambda key: f'{key} < ?',
	GREATER = lambda key: f'{key} > ?',
	LESS_EQUAL = lambda key: f'{key} <= ?',
	GREATER_EQUAL = lambda key: f'{key} >= ?',
	EQUAL = lambda key: f'{key} = ?',
	NOT_EQUAL = lambda key: f'{key} != ?',
	IN = lambda key: f'{key} IN (?)',
	NOT_IN = lambda key: f'{key} NOT IN (?)',
	LIKE = lambda key: f'{key} LIKE ?',
	NOT_LIKE = lambda key: f'{key} NOT LIKE ?'
)


class Statement:
	def __init__(self, table):
		self.table = table
		self.values = []

		self._where = ''
		self._order = None
		self._limit = None
		self._offset = None


	def __str__(self):
		return self.compile('sqlite')


	def where(self, key, value, comparison='equal', operator='and'):
		try:
			comp = Comparison[comparison.upper().replace('-', '_')]

		except KeyError:
			raise KeyError(f'Invalid comparison: {comparison}')

		prefix = f' {operator} ' if self._where else ' '

		if key == 'order':
			key = f"'{key}'"

		self._where += f'{prefix}{comp(key)}'
		self.values.append(value)

		return self


	def order(self, column, direction='ASC'):
		direction = direction.upper()
		assert direction in ['ASC', 'DESC']
		self._order = (column, direction)
		return self


	def limit(self, limit_num):
		self._limit = int(limit_num)
		return self


	def offset(self, offset_num):
		self._offset = int(offset_num)
		return self


	def compile(self, dbtype):
		raise NotImplementedError('Do not use the Statement class directly.')


class Select(Statement):
	def __init__(self, table, *columns, case_sensitive=True, **kwargs):
		super().__init__(table)

		self.columns = columns
		self.case_sensitive = True

		for key, value in kwargs.items():
			self.where(key, value)


	def compile(self, dbtype):
		data = f'SELECT'

		if self.columns:
			columns = ','.join(self.columns)

		else:
			columns = '*'
	
		data += f' {columns} FROM {self.table}'

		if self._where:
			data += f' WHERE {self._where}'

		if not self.case_sensitive:
			data += ' COLLATE NOCASE'

		if self._order:
			col, direc = self._order
			data += f' ORDER BY "{col}" {direc}'

		if self._limit:
			data += f' LIMIT {self._limit}'

		if self._offset:
			data += f' OFFSET {self._offset}'

		return data


class Insert(Statement):
	def __init__(self, table, **kwargs):
		super().__init__(table)

		self.keys = []

		for pair in kwargs.items():
			self.add_data(*pair)


	def add_data(self, key, value):
		self.keys.append(key)
		self.values.append(value)


	def remove_data(self, key):
		index = self.keys.index(key)

		del self.keys[index]
		del self.values[index]


	def compile(self, dbtype):
		keys = ','.join([f'"{key}"' for key in self.keys])
		values = ','.join('?' for value in self.values)
		data = f'INSERT INTO {self.table} ({keys}) VALUES ({values})'

		if dbtype == 'postgresql':
			data += ' RETURNING *'

		elif dbtype == 'sqlite' and SQLITE_VERSION >= (3, 35, 0):
			data += ' RETURNING *'

		return data + ';'


class Update(Statement):
	def __init__(self, table, **kwargs):
		super().__init__(table)
		self.keys = []

		for key, value in kwargs.items():
			self.keys.append(key)
			self.values.append(value)


	def compile(self, dbtype):
		pairs = ','.join(f'"{key}" = ?' for key in self.keys)
		data = f'UPDATE {self.table} SET {pairs} WHERE {self._where}'

		if dbtype == 'postgresql':
			data += ' RETURNING *'

		elif dbtype == 'sqlite' and SQLITE_VERSION >= (3, 35, 0):
			data += ' RETURNING *'

		return data + ';'


class Delete(Statement):
	def __init__(self, table, **kwargs):
		super().__init__(table)

		for key, value in kwargs.items():
			self.where(key, value)


	def compile(self, dbtype):
		data = f'DELETE FROM {self.table}'

		if self._where:
			data += f' WHERE {self._where}'

		if dbtype == 'postgresql':
			data += ' RETURNING *'

		elif dbtype == 'sqlite' and SQLITE_VERSION >= (3, 35, 0):
			data += ' RETURNING *'

		return data + ';'


class Count(Statement):
	def __init__(self, table, **kwargs):
		super().__init__(table)

		for key, value in kwargs.items():
			self.where(key, value)


	def compile(self, dbtype):
		data = f'SELECT COUNT(*) FROM {self.table}'

		if self._where:
			data += f' WHERE {self._where}'

		return data


class AlterTable(Statement):
	def compile(self, sql):
		return f"ALTER TABLE '{self.table}' {sql}"


	def rename(self, new_name):
		return self.compile(f"RENAME TO '{new_name}'")


	def rename_column(table, column, new_name):
		return self.compile(f"RENAME COLUMN '{column}' TO '{new_name}'")


	def add_column(table, *args, dbtype='sqlite', **kwargs):
		col = DbColumn(dbtype, *args, **kwargs)

		if col.nullable and not col.default:
			raise ValueError('A default value must be provided for nullable columns')

		if col.foreign_key and col.default != None:
			raise ValueError('A default value of "None" must be used if a foreign key is set')

		return self.compile(f"ADD COLUMN {col.compile()}")


	def drop_column(table, column):
		return self.compile(f"DROP TABLE '{column}'")
