from izzylib import DotDict, LruCache, izzylog
from izzylib.object_base import ObjectBase


class SessionTables:
	def __init__(self, session):
		self._session = session
		self._db = session.db
		self._tables = session.db.tables


	#def __getattr__(self, key):
		#return SessionTable(self.session, key, self._tables[key])


	def __getitem__(self, key):
		if not self._tables.get(key):
			raise KeyError(key)
		return SessionTable(self._session, key, self._tables[key])


	def get(self, name):
		try:
			return self[name]

		except KeyError:
			return None


	def names(self):
		return tuple(self._tables.keys())


	def keys(self):
		return self._tables.keys()


class SessionTable(DotDict):
	def __init__(self, session, name, columns):
		super().__init__(columns)

		self._name = name
		self._session = session
		self._db = session.db


	@property
	def name(self):
		return self._name


	@property
	def columns(self):
		return tuple(self.keys())


	def fetch(self, *args, **kwargs):
		#self._check_columns(**kwargs)
		return self.session.fetch(self.name, *args, **kwargs)


	def insert(self, *args, **kwargs):
		#self._check_columns(**kwargs)
		return self.session.insert(self.name, *args, **kwargs)


	def remove(self, *args, **kwargs):
		#self._check_columns(**kwargs)
		return self.session.remove(self.name, *args, **kwargs)


	def _check_columns(self, **kwargs):
		for key in kwargs.keys():
			if key not in self.columns:
				raise KeyError(f'Not a column for table "{self.name}": {key}')


class DbTables(DotDict):
	def __init__(self, db):
		super().__init__()

		self._db = db


	@property
	def empty(self):
		return len(self.keys()) == 0


	@property
	def names(self):
		return tuple(self.keys())


	def load_tables(self, tables):
		for name, columns in tables.items():
			self.add_table(name, columns)


	def unload_tables(self):
		for key in self.names:
			self.remove_table(name)

		self._db.cache.clear()


	def add_table(self, name, columns):
		izzylog.verbose('Adding table:', name)

		self[name] = {}

		if type(columns) == list:
			columns = {col.name: col for col in columns}

		for column, data in columns.items():
			self.add_column(name, column, data)

		self._db.cache[name] = LruCache()


	def remove_table(self, name):
		izzylog.verbose('Removing table:', name)

		self._db.cache.pop(name)
		return self.pop(name)


	def get_columns(self, name):
		return tuple(self[name].values())


	def add_column(self, table, name, data):
		if type(data) == Column:
			data = DbColumn.new_from_column(self._db.cfg.type, data)

		elif type(data) == dict:
			data = DbColumn(self._db.cfg.type, name, **data)

		elif type(data) == DbColumn:
			pass

		else:
			raise TypeError(f'Invalid column data type: {type(data).__name__}')

		if not self.get(table):
			self[table] = {}

		self[table][name] = data


	def compile_table(self, table_name, dbtype):
		table = self[table_name]
		columns = []
		foreign_keys = []

		for column in self.get_columns(table_name):
			columns.append(column.compile(dbtype))

			if column.foreign_key:
				fkey_table, fkey_col = column.foreign_key
				foreign_keys.append(f"FOREIGN KEY ('{column.name}') REFERENCES '{fkey_table}' ('{fkey_col}')")

		if foreign_keys:
			return f"CREATE TABLE IF NOT EXISTS '{table_name}' ({','.join(columns)}, {','.join(foreign_keys)})"

		else:
			return f"CREATE TABLE IF NOT EXISTS '{table_name}' ({','.join(columns)})"


	def compile_all(self, dbtype):
		return [self.compile_table(name, dbtype) for name in self.keys()]


class DbColumn(DotDict):
	def __init__(self, dbtype, name, type=None, default=None, primary_key=False, unique=False, nullable=True, autoincrement=False, foreign_key=None):
		super().__init__(
			name = name,
			type = type,
			default = default,
			primary_key = primary_key,
			unique = unique,
			nullable = nullable,
			autoincrement = autoincrement,
			foreign_key = foreign_key
		)

		self._dbtype = dbtype

		if self.name == 'id':
			if dbtype == 'sqlite':
				self.type = 'INTEGER'
				self.autoincrement = True

			elif dbtype == 'postgresql':
				self.type = 'SERIAL'
				self.autoincrement = False

			self.primary_key = True
			self.unique = False
			self.nullable = True
			self.default = None
			self.foreign_key = None

		elif self.name in ['created', 'modified', 'accessed'] and not self.type:
			self.type = 'DATETIME'

		if not self.type:
			raise ValueError(f'Must provide a column type for column: {name}')

		try:
			self.fkey

		except ValueError:
			raise ValueError(f'Invalid foreign_key format. Must be "table.column"')


	@classmethod
	def new_from_column(cls, dbtype, column):
		return cls(dbtype, column.name,
			type=column.type,
			default=column.default,
			primary_key=column.primary_key,
			unique=column.unique,
			nullable=column.nullable,
			autoincrement=column.autoincrement,
			foreign_key=column.foreign_key
		)


	@property
	def fkey(self):
		try:
			return self.foreign_key.split('.')

		except AttributeError:
			return


	def compile(self, *args):
		line = f"'{self.name}' {self.type}"

		if self.primary_key:
			line += ' PRIMARY KEY'

		if not self.nullable:
			line += ' NOT NULL'

		if self.unique:
			line += ' UNIQUE'

		if self.autoincrement and self._dbtype != 'postgresql':
			line += ' AUTOINCREMENT'

		if self.default:
			line += f" DEFAULT {parse_default(self.default)}"

		return line


class Column(DotDict):
	def __init__(self, name, type=None, default=None, primary_key=False, unique=False, nullable=True, autoincrement=False, foreign_key=None):
		super().__init__(
			name = name,
			type = type,
			default = default,
			primary_key = primary_key,
			unique = unique,
			nullable = nullable,
			autoincrement = autoincrement,
			foreign_key = foreign_key
		)

		if self.name == 'id':
			self.type = 'SERIAL'

		elif self.name in ['created', 'modified', 'accessed'] and not self.type:
			self.type = 'DATETIME'

		if not self.type:
			raise ValueError(f'Must provide a column type for column: {name}')


	def __setitem__(self, key, value):
		if key == 'type' and value:
			value = value.upper()

		if key == 'foreign_key' and value:
			try: table, column = value.split('.', 1)
			except ValueError: raise ValueError('Invalid foreign_key format. Must be "table.column"')

			value = (table, column)

		DotDict.__setitem__(self, key, value)


def parse_default(default):
	if isinstance(default, dict) or isinstance(default, list):
		default = json.dumps(default)

	if type(default) == str:
		default = f"'{default}'"

	return default
