from datetime import date, time, datetime
from izzylib import DotDict, LowerDotDict, class_name, izzylog


Standard = {
	'INTEGER',
	'INT',
	'TINYINT',
	'SMALLINT',
	'MEDIUMINT',
	'BIGINT',
	'UNSIGNED BIG INT',
	'INT2',
	'INT8',
	'TEXT',
	'CHARACTER',
	'CHAR',
	'VARCHAR',
	'BLOB',
	'CLOB',
	'REAL',
	'DOUBLE',
	'DOUBLE PRECISION',
	'FLOAT',
	'NUMERIC',
	'DEC',
	'DECIMAL',
	'BOOLEAN',
	'DATE',
	'TIME',
	'JSON'
}


Sqlite = {
	*Standard,
	'DATETIME'
}


Postgresql = {
	*Standard,
	'SMALLSERIAL',
	'SERIAL',
	'BIGSERIAL',
	'VARYING',
	'BYTEA',
	'TIMESTAMP',
	'INTERVAL',
	'POINT',
	'LINE',
	'LSEG',
	'BOX',
	'PATH',
	'POLYGON',
	'CIRCLE',
}


Mysql = {
	*Standard,
	'FIXED',
	'BIT',
	'YEAR',
	'VARBINARY',
	'ENUM',
	'SET'
}


class Type:
	sqlite = None
	postgresql = None
	mysql = None


	def __getitem__(self, key):
		if key in ['sqlite', 'postgresql', 'mysql']:
			return getattr(self, key)

		raise KeyError(f'Invalid database type: {key}')


	def __call__(self, action, dbtype, value):
		assert action in ['serialize', 'deserialize']
		return getattr(self, action)(dbtype, value)


	def name(self, dbtype='sqlite'):
		return self[dbtype]


	def serialize(self, dbtype, value):
		return value


	def deserialize(self, dbtype, value):
		return value


class Json(Type):
	sqlite = 'JSON'
	postgresql = 'JSON'
	mysql = 'JSON'


	def deserialize(self, dbtype, value):
		izzylog.debug(f'deserialize {class_name(self)}: {class_name(value)}', value)

		if type(value) == str:
			return DotDict(value)

		return value


	def serialize(self, dbtype, value):
		izzylog.debug(f'serialize {class_name(self)}: {class_name(value)}', value)

		if not isinstance(value, DotDict):
			value = DotDict(value)

		return value.to_json()


class Datetime(Type):
	sqlite = 'DATETIME'
	postgresql = 'TIMESTAMP'
	mysql = 'DATETIME'


	def deserialize(self, dbtype, value):
		izzylog.debug(f'deserialize {class_name(self)}: {class_name(value)}', value)

		if type(value) == str:
			return datetime.fromisoformat(value)

		elif type(value) == int:
			return datetime.fromtimestamp(value)

		return value


	def serialize(self, dbtype, value):
		izzylog.debug(f'serialize {class_name(self)}: {class_name(value)}', value)

		if dbtype == 'sqlite':
			return value.isoformat()

		return value


class Date(Type):
	sqlite = 'DATE'
	postgresql = 'DATE'
	mysql = 'DATE'


	def deserialize(self, dbtype, value):
		izzylog.debug(f'deserialize {class_name(self)}: {class_name(value)}', value)

		if type(value) == str:
			return date.fromisoformat(value)

		elif type(value) == int:
			return date.fromtimestamp(value)


		return value


	def serialize(self, dbtype, value):
		izzylog.debug(f'serialize {class_name(self)}: {class_name(value)}', value)

		if dbtype == 'sqlite':
			return value.isoformat()

		return value


class Time(Type):
	sqlite = 'TIME'
	postgresql = 'TIME'
	mysql = 'TIME'


	def deserialize(self, dbtype, value):
		izzylog.debug(f'deserialize {class_name(self)}: {class_name(value)}', value)

		if type(value) == str:
			return time.fromisoformat(value)

		elif type(value) == int:
			return time.fromtimestamp(value)

		return value


	def serialize(self, dbtype, value):
		izzylog.debug(f'serialize {class_name(self)}: {class_name(value)}', value)

		if dbtype == 'sqlite':
			return value.isoformat()

		return value


class Types(DotDict):
	def __init__(self, db):
		self._db = db


	def _setup_types(self):
		self.set_type(Json, Date, Time, Datetime)


	def get_type(self, name):
		return self.get(name.upper(), Type())


	def set_type(self, *types):
		for type_object in types:
			typeclass = type_object()
			self[typeclass.name(self._db.cfg.type)] = typeclass
